/********************************************************************************
	Copyright 2005 Sjaak Priester	

	This is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this software; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

	http://www.gnu.org/licenses/
********************************************************************************/

// QBlockText
//
// C++ class to fill a Windows GDI+ RectF with lines of text.
//
// The resulting text is assembled in another GDI+ GraphicsPath by calling
// the member function GetBlockText().
// -------------------------------
//
// QEnvelopeText
//
// C++ class to fill a Windows GDI+ path (the envelope path) with lines of text.
//
// The resulting text is assembled in another GDI+ GraphicsPath by calling
// the member function GetEnvelopeText().
// -------------------------------
//
// See the QEnvelopeTextDemo application for a demonstration.
//
// Source:
//		QEnvelopeText.cpp
//
// Other files needed:
//		QGraphicsText.h
//		QGraphicsText.cpp
//
// Environment: Windows 98 and later, GDI+, Visual C++ 7.0/7.1, MFC.
// Probably compatible with Visual C++ 6.0 as well.
//
//==============
// Version 1.0, September 21, 2005.
// (c) 2005, Sjaak Priester, Amsterdam.
// http://www.sjaakpriester.nl
//
// Use at your ownn risk. Comments welcome.

#pragma once

#include <map>
#include <vector>
using namespace std;

#include "QGraphicsText.h"

class QBlockText : public QGraphicsText
{
public:
	QBlockText();
	virtual ~QBlockText();

	int GetBlockText(GraphicsPath& pathOutput, LPCTSTR str, const RectF& rect) const;
	// Add the text string str to the GraphicsPath pathOutput.
	// The line of text will be laid out in the enclosing RectF rect.
	//
	// If str is too long to fit completely inside the rectangle, it will be truncated.
	//
	// Return: number of characters processed (may be less than length of str).
	//
	// Note that GetEnvelopeText will not clear pathOutput, before it adds the string.
	// It will set pathOutput's fill mode to FillModeWinding.

	bool SetScaleMargins(REAL min, REAL max);
	void GetScaleMargins(REAL& min, REAL& max) const;
	// These two values determine the amount by which an individual text line may be
	// horizontally shrinked or expanded to make it fit inside the enclosing rectangle.
	//		min must have a value between 0 and 1 (default: 0.9f);
	//		max must be greater than 1 (default: 1.1f).
	// If SetScaleMargins() returns false, min, or max, or both, are out of range.

	enum justification
	{
		justLeft,
		justCenter,
		justRight,
		justFull,
		justHard
	};
	void SetJustification(justification Just)		{ m_Justification = Just; }
	justification GetJustification() const			{ return m_Justification; }
	// Set or get the justification mode.
	//		justFull breaks line only at valid positions; this means that text may exceed the enclosing
	//			rectangle if no valid position exists.
	//		justHard is the same as justFull, except words will be broken right in the middle,
	//			if no other options are available.

	TCHAR m_Hyphen;
	// The hyphen character, used if GetBreakType() returns brkHyphen. Change this if you like.

protected:
	enum brkType
	{
		fBreak = 0x0001,	// flags
		fScale = 0x0002,
		fSkip = 0x0004,
		fDash = 0x0008,

		brkNone = 0,							// Do not break text here, if possible (normal characters).
		brkNormal = fBreak | fScale,			// Text break allowed here.
		brkSpace = fBreak | fScale | fSkip,		// Text break allowed here; if applied, skip this character (white space).
		brkHyphen = fBreak | fScale | fDash,	// Text break allowed here; if applied, insert hyphen.
		brkReturn = fBreak | fSkip,				// Do break text here, and skip this character (return).
		brkHard = fBreak						// Do break text here ('hard' break, in the middle of word).
	};

	virtual UINT GetBreakType(LPCTSTR str, int index, int len) const;
	// Overridable by derived class.
	// Return one of the brkXxx values of brkType.
	//
	// Default version returns:
	//		brkReturn		if str[index] == '\r';
	//		brkSpace		if isspace(c) is true, where c = str[index];
	//		brkNormal		if str[index - 1] == '-' ('hard hyphen');
	//		brkNone			otherwise.
	// The default version never returns brkHyphen or brkHard, but overrides may do so.

	struct Strip
	{
		Strip(REAL x, REAL y, REAL len)
			: m_x(x), m_y(y), m_len(len)	{ }

		REAL m_x;
		REAL m_y;
		REAL m_len;
	};

	typedef vector<Strip> Strips;

	int OutputText(GraphicsPath& pathOutput, LPCTSTR str, const Strips& strips) const;

	REAL m_ScaleMin;
	REAL m_ScaleMax;
	justification m_Justification;
};

class QEnvelopeText : public QBlockText
{
public:
	QEnvelopeText(const GraphicsPath& pathEnvelope, REAL flatness = 10.0f * FlatnessDefault);
	// Create a QEnvelopeText with pathEnvelope as enclosing figure.
	// pathEnvelope may consist of one or more closed subpaths.
	// Parameter flatness determines the accuracy of the 'flattening'. The default
	// will be appropriate for almost any case.

	virtual ~QEnvelopeText();

	int GetEnvelopeText(GraphicsPath& pathOutput, LPCTSTR str) const;
	// Add the string str to the GraphicsPath pathOutput.
	// The string will be laid out in the enclosing envelope figure.
	//
	// If str is too long to fit completely inside the envelope path, it will be truncated.
	//
	// Return: number of characters processed (may be less than length of str).
	//
	// Note that GetEnvelopeText will not clear pathOutput, before it adds the string.
	// It will set pathOutput's fill mode to FillModeWinding.

protected:
	typedef multimap<REAL, int> HorScan;

	void GetXPoint(HorScan& xpoints, REAL y, int lineType, const PointF& p0, const PointF& p1) const;
	void GetXPoints(HorScan& xpoints, REAL y, int lineType) const;

	GraphicsPath * m_pEnvelope;
};
